/* $Id: hawki_cal_lingain.c,v 1.4 2010/09/28 14:13:48 cgarcia Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: cgarcia $
 * $Date: 2010/09/28 14:13:48 $
 * $Revision: 1.4 $
 * $Name: hawki-1_8_0 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------
                             Includes and Defines
 ----------------------------------------------------------------------------*/

#include <cpl.h>

#include "irplib_detmon_lg.h"
#include "irplib_detmon.h"
#include "irplib_plugin.h"

#include "hawki_dfs.h"

#define RECIPE_NAME "hawki_cal_lingain"

/* Copy here instrument specific keywords which need to be in the PAF file */
#define INSTREGEXP   "ESO INS SETUP ID"
#define PAFREGEXP    "^(" REGEXP "|" INSTREGEXP ")$"


/*----------------------------------------------------------------------------
			     Functions prototypes
 ----------------------------------------------------------------------------*/

/**@{*/

cpl_error_code hawki_cal_lingain_fill_parlist_default(cpl_parameterlist * parlist);

IRPLIB_RECIPE_DEFINE(hawki_cal_lingain, HAWKI_BINARY_VERSION,
        hawki_cal_lingain_fill_parlist_default(recipe->parameters),
        "Enrique Garcia", PACKAGE_BUGREPORT, "2009", 
        "Linearity/Gain recipe for the IR domain",
        irplib_detmon_lg_get_description(RECIPE_NAME, "HAWKI", 
                                         HAWKI_CAL_LINGAIN_LAMP_RAW,
                                         HAWKI_CAL_LINGAIN_DARK_RAW));

/*---------------------------------------------------------------------------*/
/*
  @brief    Interpret the command line options and execute the data processing
  @param    frameset    the frames list
  @param    parlist     the parameters list
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/
static int hawki_cal_lingain(cpl_frameset            * frameset,
                             const cpl_parameterlist * parlist)
{
    cpl_error_code        error ;
    cpl_propertylist  *   pro_lintbl ;
    cpl_propertylist  *   pro_gaintbl ;
    cpl_propertylist  *   pro_coeffscube ;
    cpl_propertylist  *   pro_bpm ;
    cpl_propertylist  *   pro_corr ;
    cpl_propertylist  *   pro_diff ;
    
    /* Create the PRO keys propertylist */
    pro_lintbl = cpl_propertylist_new() ;
    cpl_propertylist_append_string(pro_lintbl, CPL_DFS_PRO_CATG, 
            HAWKI_CALPRO_LINGAIN_LIN) ;
    cpl_propertylist_append_string(pro_lintbl, CPL_DFS_PRO_TYPE,
            HAWKI_PROTYPE_LIN_STATS) ;
    
    pro_gaintbl = cpl_propertylist_new() ;
    cpl_propertylist_append_string(pro_gaintbl, CPL_DFS_PRO_CATG, 
            HAWKI_CALPRO_LINGAIN_GAIN) ;
    cpl_propertylist_append_string(pro_gaintbl, CPL_DFS_PRO_TYPE,
            HAWKI_PROTYPE_GAIN_STATS) ;
    
    pro_coeffscube = cpl_propertylist_new() ;
    cpl_propertylist_append_string(pro_coeffscube, CPL_DFS_PRO_CATG, 
            HAWKI_CALPRO_LINGAIN_COEFFS) ;
    cpl_propertylist_append_string(pro_coeffscube, CPL_DFS_PRO_TYPE, 
            HAWKI_PROTYPE_LIN_COEFFS) ;
    
    pro_bpm = cpl_propertylist_new() ;
    cpl_propertylist_append_string(pro_bpm, CPL_DFS_PRO_CATG, 
            HAWKI_CALPRO_LINGAIN_BPM) ;
    cpl_propertylist_append_string(pro_bpm, CPL_DFS_PRO_TYPE, 
            HAWKI_PROTYPE_BPM_LEVELS) ;
    
    pro_corr = cpl_propertylist_new() ;
    cpl_propertylist_append_string(pro_corr, CPL_DFS_PRO_CATG, 
            HAWKI_CALPRO_LINGAIN_CORR) ;
    cpl_propertylist_append_string(pro_corr, CPL_DFS_PRO_TYPE,
            HAWKI_PROTYPE_LINGAIN_CORR) ;
    
    pro_diff = cpl_propertylist_new() ;
    cpl_propertylist_append_string(pro_diff, CPL_DFS_PRO_CATG, 
            HAWKI_CALPRO_LINGAIN_DIFF) ;
    cpl_propertylist_append_string(pro_diff, CPL_DFS_PRO_TYPE,
            HAWKI_PROTYPE_LINGAIN_DIFF) ;

    /* Call the lingain function */
    error = irplib_detmon_lg(frameset,
            parlist,
            HAWKI_CAL_LINGAIN_LAMP_RAW,
            HAWKI_CAL_LINGAIN_DARK_RAW,
            RECIPE_NAME,
            PACKAGE_TARNAME,
            PAFREGEXP,
            pro_lintbl, 
            pro_gaintbl, 
            pro_coeffscube, 
            pro_bpm, 
            pro_corr, 
            pro_diff,
            PACKAGE "/" PACKAGE_VERSION,
            NULL, NULL, TRUE);

    /* Delete the PRO keys propertylist */
    cpl_propertylist_delete(pro_lintbl) ;
    cpl_propertylist_delete(pro_gaintbl) ;
    cpl_propertylist_delete(pro_coeffscube) ;
    cpl_propertylist_delete(pro_bpm) ;
    cpl_propertylist_delete(pro_corr) ;
    cpl_propertylist_delete(pro_diff) ;

    /* Propagate the error, if any */
    cpl_ensure_code(!error, error);

    /* Return */
    if (cpl_error_get_code())
    {
        cpl_msg_error(__func__,
                      "HAWK-I pipeline could not recover from previous errors");
        return -1 ;
    }
    else return 0;
}

cpl_error_code
hawki_cal_lingain_fill_parlist_default(cpl_parameterlist * parlist)
{
    cpl_parameter   *   p;
    cpl_error_code      error ;
        
    error=irplib_detmon_lg_fill_parlist_nir_default(parlist,
            RECIPE_NAME, PACKAGE_TARNAME);
    cpl_ensure_code(!error, error);

    /* Set to reduce all extensions */
    p = cpl_parameterlist_find(parlist, PACKAGE_TARNAME "." RECIPE_NAME 
            ".exts");
    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
    error = cpl_parameter_set_default_int(p, -1);
    cpl_ensure_code(!error, error);


    return CPL_ERROR_NONE;
}
